#ifndef lint
static char sccsid[] = "@(#)uns_srvr.c 7.9 92/09/09 SMI";
#endif

/*
 * Copyright (c) 1987 by Sun Microsystems, Inc.
 */

/* START_LEGAL_STUFF	:	cpywr-1989v1.0 */
/*
 *************************************************************************
 *	Copyright (c) 1989, Sun Microsystems, Inc.  All Rights Reserved.
 *	Sun considers its source code as an unpublished, proprietary
 *	trade secret, and it is available only under strict license
 *	provisions.  This copyright notice is placed here only to protect
 *	Sun in the event the source is deemed a published work.  Dissassembly,
 *	decompilation, or other means of reducing the object code to human
 *	readable form is prohibited by the license agreement under which
 *	this code is provided to the user or company in possession of this
 *	copy.
 *
 *	RESTRICTED RIGHTS LEGEND: Use, duplication, or disclosure by the
 *	Government is subject to restrictions as set forth in subparagraph 
 *	(c)(1)(ii) of the Rights in Technical Data and Computer Software
 *	clause at DFARS 52.227-7013 and in similar clauses in the FAR and
 *	NASA FAR Supplement.
 *************************************************************************
 */
/* END_LEGAL_STUFF	:	cpywr-1989v1.0 */

 
/*
Module: 	uns_srvr.c
  
 ******************************************************************************
 *  COPYRIGHT (C) 1985, 1986 by TECHNOLOGY CONCEPTS INC.                      *
 *                              SUDBURY, MASSACHUSETTS 01776                  *
 *                              ---ALL RIGHTS RESERVED---                     *
 *                                                                            *
 *  THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED     *
 *  ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH THE INCLUSION  *
 *  OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY OTHER COPIES THEREOF *
 *  MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY OTHER PERSON.  NO  *
 *  TITLE TO AND OWNERSHIP OF THE SOFTWARE IS HEREBY TRANSFERRED.             *
 *                                                                            *
 *  THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE AND  *
 *  SHOULD NOT BE CONSTRUED AS A COMMITMENT BY TECHNOLOGY CONCEPTS INC.       *
 ******************************************************************************
  
Version 1  --  Revision 0
  
Author: 		Brad Karasik
			Brian Anderson
Routines:
	get_new_server(fd, a_info)
	change_uid(ai)
	log(reason, ai)	
	lookup(number, name, file)
	objmatch(o_number,f_number,o_name,f_name)
	server_name(string)
	str_lc(string)
*/

#define EXISTS        0
#define EXECUTABLE    1

#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <pwd.h>
#include <sys/ioctl.h>
#include <sys/signal.h>
#include <sys/time.h>
#include <netdni/tcistd.h>
#include <netdni/dni.h>
#include <errno.h>
#include <pwd.h>

extern int errno;

extern char *lookup();		/* to find the record in the registry */
extern short change_uid();	/* to change the uid and set up environs */

Char *server_name();		/* gives filename of the servers path */
Short objmatch();		/* routine used by lookup to check objects */

extern Char *server_table;
extern Char *server_log;
extern Char *users_servers_log;
extern Char *child_home;

char user[20];
int ctermsrv = 0;

/*
 *	Routine :		get_new_server()
 *
 *	Abstract:		This routine will do one of two things. It
 *				will attempt to locate and start the appropriate
 *				server if possible. If it can find the server
 *				that the client wants to connect to it will 
 *				exec that server. If it can't find it or access
 *			 	is denied, the routine will return a bad connect
 *				response to the board, and will then exit.
 *
 *	Explicit Inputs:	int fd;
 *				OpenBlock *a_info;
 */
void
get_new_server(fd, a_info)
	int fd;
	OpenBlock *a_info;
{
	char *server;
	char object[100];
	SessionData ses_data;
	int i;
	char filedes[20];
	int retval;
	FILE *fpt;
	long current_time;
	char *getenv();
	char *ncp_dir;

	/* make the userid lowercase std on UNIX */
	str_lc(a_info->op_userid);
	/* keep the name of the user */
	strcpy(user, a_info->op_userid);

	/*
	 * lets try to change this process to that of the user who is doing
	 * the connect 
	 */
	ctermsrv = 0;
	if ((retval = change_uid(a_info)) == FAIL) {
		/* bad user id */
		ses_data.sd_reason = ACCESS_CONT;

		ses_data.sd_data.im_length = 0;
		ioctl(fd, SES_REJECT, &ses_data);
		close(fd);
		exit(1);
	}

	/*
	 * User has been logged in and we are running under his id and
	 * allegedly in his default login directory 
	 */

	/*
	 * check to see if the openblock is formed correctly... in the past
	 * it could be wrong..... ie BOTH the object number and name were
	 * both defined...it SHOULD be an either or field 
	 */
	if ((a_info->op_object_nbr != 0)
	    && (a_info->op_task_name[0] != 0)) {
		log("BAD OpenBlock", 0);
		ses_data.sd_reason = BAD_OBJECT;
		ses_data.sd_data.im_length = 0;
		ioctl(fd, SES_REJECT, &ses_data);
		close(fd);
		exit(1);
	}

	/*
	 * put the object number into string form for lookup 
	 */
	sprintf(object, "%d", a_info->op_object_nbr);

	/*
	 * save the logical link filedes so we can send it to our exec'ed
	 * process 
	 */
	sprintf(filedes, "%d", fd);

	/* if NCP databases not in default location, add to env_list */
	ncp_dir = getenv("DNI_NCP_DIR");
	if (ncp_dir)
		env_build("DNI_NCP_DIR", ncp_dir);
	/*
	 * look for the task in the system table by number or name 
	 */
	server = lookup(object, a_info->op_task_name, "");
	if (server != NULL) {
		/*
		 * the task exists! log the connect and exec it. 
		 */

		/* first to the system servers log */
		log("SRVR OK", a_info);
		fclose(stderr);

		if (!ctermsrv) {
			fopen(users_servers_log, "a");

			/* and again to the users log */
			log("SRVR OK", a_info);
		}

		execle(server,
		       server_name(server), filedes, 0,
		       env_head());

		/*
		 * Only the bad execs gets here.  Normal execle
		 * calls NEVER reach this stretch of code.
		 */
		log("SERVER EXECLE failed", 0);

		/* reject it as bad object... at least on this system */
		ses_data.sd_data.im_length = 0;
		ses_data.sd_reason = BAD_OBJECT;
		ioctl(fd, SES_REJECT, &ses_data);
		close(fd);
		exit(1);
	}

	/*
	 * NOT in the system servers table check if the task exists in the
	 * users directory and if it is an executable file BUT ONLY if it is
	 * a task name and not just a number we dont need to check vs number
	 * since due to earlier check to see if both name and number 
	 */
	if (a_info->op_task_name[0] != 0) {
		if (access(a_info->op_task_name, EXECUTABLE) == 0) {
			/*
			 * task_name is executable by the user
			 * log it and execute it
			 *
			 * log once to the server log 
			 */
			log("UPROC OK", a_info);

			fclose(stderr);
			fopen(users_servers_log, "a");

			/* log again to the users log */
			log("UPROC OK", a_info);
			execle(a_info->op_task_name,
			       server_name(a_info->op_task_name), filedes, 0,
			       env_head());
			log("UPROC EXEC failed", 0);
		} else {
			/* cant execute task */
			log("BAD UPROC", a_info);
			ses_data.sd_reason = BAD_OBJECT;
		}
	} else {
		/*
		 * Specified object/task is neither in server table nor is it
		 * in the users login directory. 
		 */
		log("NO SERVER", a_info);
		ses_data.sd_reason = BAD_OBJECT;
	}

	/*
	 * Only the bad execs and files not found can get here normal execls
	 * NEVER reach this stretch of code 
	 */

	ses_data.sd_data.im_length = 0;
	ioctl(fd, SES_REJECT, &ses_data);
	close(fd);
	exit(1);
}

/*
 *
 *	Routine :		change_uid();
 *
 *	Abstract:		This routine attempts to change the process
 *				to be that of the connector of the link. It
 *				checks the password and if successful will
 *				change the current directory to be the home
 *				directory of the connector. It also changes
 *				the group id
 *
 *	Explicit Inputs:	OpenBlock *ai;
 *
 *	Return Value:		FAIL	- couldnt do it.
 *				SUCCESS - process has been changed to that
 *					  of the connector.
 *
 *	Side Effects:		the process is now owned by the connector
 *				and the default directory is that of the user.
 */
short
change_uid(ai)
	OpenBlock *ai;
{
	struct passwd *p;
	char *namep;
	struct passwd *getpwnam();
	char *crypt();
	FILE *fp3;
	Tbool bypass = FALSE;
	extern Char null_user[];
	extern Char priv_user[];
	short get_proxy();
	struct stat buf;

	if (ai->op_task_name == "CTERM" || ai->op_task_name == "cterm"
				|| ai->op_object_nbr == 42) {
		strcpy(ai->op_userid, "root");
		ctermsrv = 1;
		bypass = TRUE;
	}
	/* if the userid is an empty string use defaults */
	/* first try for proxy access	*/
	else if ((ai->op_userid[0] == 0) || (ai->op_userid[0] == ' ')) {
		if (get_proxy(ai) == NULL) {
			get_priv();
			strcpy(ai->op_userid, null_user);
		}
		bypass = TRUE;	/* bypass the password checking */
	}

	/* is there such a user ? */
	if ((p = getpwnam(ai->op_userid)) == NULL) {
		log("BAD UID", ai);
		return FAIL;
	}

	/*
	 * Root or anyuser that has the uid of 0 which is the superuser is
	 * not allowed to use this facility.  This is a matter of the most
	 * basic of security rules.  The system administrator may declare
	 * a priv_user via the get_priv call and priv_file. 
	 */
	if (p->pw_uid == 0 && !ctermsrv) {
		get_priv();
		if (strcmp(ai->op_userid, priv_user) != 0) {
			log("REMOTE ROOT", ai);
			return FAIL;
		}
	}

	/* if this user does exist is his/her password correct? */
	if ((*(p->pw_passwd)) && (bypass != TRUE)) {
		/* test password ONLY if one is present in password file */
		namep = crypt(ai->op_password, p->pw_passwd);
		if (strcmp(namep, p->pw_passwd)) {
			/* passwords did not match */
			log("BAD PW", ai);
			return FAIL;
		}
	}

	/*
	 * dnilogind must be started as root or else tty
	 * permissions can't be reset at end of session
	 */
	if (ctermsrv) {
		if ((p = getpwnam("root")) == NULL) {
			log("BAD CT", ai);
			return(FAIL);
		}
	}

	/* set this process to be the uid, gid and dir of the user */
	/* make sure directory exists if not set to /tmp */
	fflush(stderr);
	if (stat(p->pw_dir,&buf) == -1)
		strcpy(p->pw_dir, "/tmp");
	chdir(p->pw_dir);
	/* change from setegid to setgid to set effective and real id's */
	setgid(p->pw_gid);
	initgroups(p->pw_name, p->pw_gid);	/* do all grps usr is in */
	setuid(p->pw_uid);
	child_home = p->pw_dir;	/* used in environments */
	mkenv();
	return SUCCESS;
}

/*
 * log - log a message to stderr
 *
 * Message sent to stderr, which points to dniserver.log or private user log.
 */
log(reason, ai)
	char *reason;
	OpenBlock *ai;
{
	struct tm *localtime();
	long clock;
	struct tm *utmp;
	char *name = 0;
	char *user = 0;
	char task[128];

	task[0] = 0;
	if (ai) {
		name = ai->op_node_name;
		user = ai->op_userid;
		if (ai->op_object_nbr) {
			if (ai->op_task_name[0])
				sprintf(task, "Bad Task");
			else
				sprintf(task, "%d", ai->op_object_nbr);
		} else
			sprintf(task, "%s", ai->op_task_name);
	}

	time(&clock);
	utmp = localtime(&clock);
	if (ai)
	fprintf(stderr,
"%02d/%02d %02d:%02d:%02d,node=%s,task=%s,user=%s,pid=%d,errno=%d: %s\n",
		utmp->tm_mon + 1, utmp->tm_mday,
		utmp->tm_hour, utmp->tm_min, utmp->tm_sec,
		name, task, user, getpid(), errno, reason);
	else
	fprintf(stderr,
		"%02d/%02d %02d:%02d:%02d,pid=%d,errno=%d: %s\n",
		utmp->tm_mon + 1, utmp->tm_mday,
		utmp->tm_hour, utmp->tm_min, utmp->tm_sec,
		getpid(), errno, reason);
	fflush(stderr);
}

/*
 *	Routine :		lookup
 *
 *	Abstract:		This routine looks up a server object in
 *				the file dniserver.reg and if the server
 *				exists returns the name of the file to
 *				exec. If it doesnt exist it returns NULL.
 *
 *	Explicit Inputs:	
 *				char * number;	 the object number
 *				char * name;	 the object name
 *				char * file;	 the task name
 *
 *	Return Value:		The name of the application to execute or
 *				Null if not found
 */
char *
lookup(number, name, file)
	char *number;
	char *name;
	char *file;
{
	char nambuf[128];
	char numbuf[128];
	static char filbuf[512];	/* MUST be static since we return it */
	FILE *fp;

	/* Is there a server registration file? */
	fp = fopen(server_table, "r");
	if (fp == (FILE *) NULL) {
		log("lookup can't open the server_table", 0);
		return NULL;
	}

	do {
		if (fscanf(fp, "%s %s %s\n", numbuf, nambuf, filbuf) < 0) {
			/* out of data */
			fclose(fp);
			return NULL;
		}

		/*
		 * now that we have the object number, object name and
		 * application names, lets check them against the one
		 * requested by the daemon. This loop will iterate until
		 * either a match is found or the end of file is reached. 
		 */

	} while (objmatch(number, numbuf, name, nambuf));

	/* we matched, return the name of the application to exec. */
	fclose(fp);
	return filbuf;
}

/*
 * Routine:	objmatch
 *
 * Abstract:	This routine checks the pairs of strings to determine if
 *		the given object names and numbers (o_*) match the entry
 *		from the file (f_*). If they match SUCCESS else FAIL.
 *		NOTE assumes a legal open block ie one or the other of
 *		object by number or object by name.
 *		For the moment ignore the entry about the file check that
 *		lookup used to do.
 *
 * Explicit Inputs:
 *	o_number - the open blocks object number (string)
 *	f_number - the files object number (string)
 *	o_name - the open blocks object name (string)
 *	f_name - the files object name (string)
 *
 * Return Values:
 *	SUCCESS - if this pair of data matches
 *	FAIL - if it dont.
*/
Short
objmatch(o_number, f_number, o_name, f_name)
	Char *o_number, *f_number;
	Char *o_name, *f_name;
{
	int oobj_number;
	int fobj_number;
	int atoi();

	oobj_number = atoi(o_number);

	if (oobj_number != 0) {
		/* open blocks specifies object by NUMBER ONLY */
		fobj_number = atoi(f_number);
		if (oobj_number == fobj_number)
			return SUCCESS;	/* correct found it */
		else
			return FAIL;	/* not correct bad choice */
	} else {
		/* must be object by NAME */
		return strcmp(o_name, f_name);
	}
}

/*
 * Routine:	server_name
 *
 * Abstract:	This routine finds the start of the filename.
 *		this means that if the string is "jumbo"
 *		this routine returns "jumbo".
 *		If string is '/usr/community/lib/fal'
 *		it would return "fal".
 * Explicit Input:
 *	string - the full path name to the server
 *
 * Return Values:
 *	a pointer to the file name itself ie after the last '/'
 */
Char *
server_name(string)
	Char *string;
{
	Char *ptr;
	Char *slash;

	ptr = string;
	slash = ptr - 1;	/* to guard against a non path */

	for (; *ptr; ptr++)
		if (*ptr == '/')
			slash = ptr;

	return (slash + 1);
}

/*
 * Routine:	str_lc
 *
 * Abstract:	This routine converts all characters of the given string
 *		from Upper case to lower case BUT ONLY if they are UPPER CASE
 *		All other characters are in lower case.
 *		The conversion takes place INPLACE.
 *
 * Explicit Input:
 *	string - the null terminated string that is to be converted.
 *
 * Explicit Output:
 *	string - contains the now all lower case string
 *
 * Return Code:	NONE (it always works or address errors if bad string)
 */
str_lc(string)
	Char *string;
{
	Char *ptr;

	ptr = string;

	for (ptr = string; *ptr; ptr++) {
		if ((*ptr >= 'A') && (*ptr <= 'Z'))
			*ptr = (*ptr - 'A') + 'a';
	}
}

short
get_proxy(ob)
OpenBlock *ob;
{
	Image16 dat;
	struct passwd *pass;
	FILE *fp;
char host[7],proxy[32],username[32];
	void str_lower();

	if ((fp = fopen("/etc/sunlink/dni/dni.proxy","r")) == NULL) 
		return(0);
	if (ioctl(3,SES_GET_PROXY, &dat) < 0) {
		log("Proxy ioctl failed");	
		return(0);
	}
	dat.im_data[dat.im_length] = NULL;
	while (fscanf(fp,"%s %s %s",host,proxy,username) != EOF) {
		if (host[0] == '#')
			continue;
		if (proxy[0] == '*' && username[0] == '*') {
			if ((host[0] == '*') || (!strcmp(host,ob->op_node_name))) 
				fclose(fp);
				strcpy(ob->op_userid,dat.im_data);
				return(1);
		}
		else if (!strcmp(proxy, dat.im_data)) {
			str_lower(ob->op_node_name);
			if (!strcmp(host,ob->op_node_name) || host[0] == '*') {
				fclose(fp);
				strcpy(ob->op_userid,username);
				return(1);
			}
		}
	}
	fclose(fp);
	return(0);
}

void
str_lower(ptr)
char *ptr;
{
        int x,i;

        x = strlen(ptr);
        for (i=0;i<x;i++)
                ptr[i] = tolower(ptr[i]);
}

